import os
from configparser import ConfigParser

DEFAULT_CACHE_DIR = os.path.join(os.path.expanduser("~"), '.cache', 'bincache')
DEFAULT_MAX_SIZE = 5 * 1024 * 1024 * 1024  # 5G
DEFAULT_LOG_FILE = ""
DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_STATS = False
CONFIG_FILE = 'bincache.conf'

_config = None

def parse_size(size_str):
    size_str = size_str.upper()
    units = {"B": 1, "K": 1024, "M": 1024**2, "G": 1024**3}
    for unit in units:
        if size_str.endswith(unit):
            return int(size_str[:-1]) * units[unit]
    return int(size_str)

def get_config():
    global _config
    if not _config:
        cache_dir = os.getenv('BINCACHE_DIR', DEFAULT_CACHE_DIR)
        config_file = os.path.join(cache_dir, CONFIG_FILE)
        config_params = {
            'max_size': DEFAULT_MAX_SIZE,
            'log_file': DEFAULT_LOG_FILE,
            'log_level': DEFAULT_LOG_LEVEL,
            'stats': DEFAULT_STATS,
            'temporary_dir': os.path.join(cache_dir, 'tmp'),
            'cache_dir': cache_dir
        }
        if os.path.exists(config_file):
            with open(config_file, 'r') as f:
                config_string = f"[DEFAULT]\n" + f.read()
            config = ConfigParser(allow_no_value=True)
            config.read_string(config_string)
            if config.has_option('DEFAULT', 'max_size') and config.get('DEFAULT', 'max_size') is not None:
                config_params['max_size'] = parse_size(config.get('DEFAULT', 'max_size'))
            if config.has_option('DEFAULT', 'log_file') and config.get('DEFAULT', 'log_file') is not None:
                log_file = config.get('DEFAULT', 'log_file')
                if not os.path.isabs(log_file) and not log_file.startswith('.' + os.sep):
                    log_file = os.path.join(cache_dir, log_file)
                config_params['log_file'] = log_file
            if config.has_option('DEFAULT', 'log_level') and config.get('DEFAULT', 'log_level') is not None:
                config_params['log_level'] = config.get('DEFAULT', 'log_level').upper()
            if config.has_option('DEFAULT', 'stats') and config.get('DEFAULT', 'stats') is not None:
                config_params['stats'] = config.getboolean('DEFAULT', 'stats')
            if config.has_option('DEFAULT', 'temporary_dir') and config.get('DEFAULT', 'temporary_dir') is not None:
                config_params['temporary_dir'] = config.get('DEFAULT', 'temporary_dir')
        _config = config_params
    return _config


# tests/config_test.py
import os
import tempfile
import pytest
from configparser import ConfigParser
from bincache.config import get_config, parse_size, DEFAULT_CACHE_DIR, CONFIG_FILE

from bincache import config as bincache_config

@pytest.fixture(autouse=True)
def reset_config():
    """在每个测试之前执行，用于重置config.py中的_config变量"""
    bincache_config._config = None

def test_parse_size():
    """测试 parse_size 函数，确保其能正确解析带有单位的字符串并转化为字节大小"""
    assert parse_size("10B") == 10
    assert parse_size("1K") == 1024
    assert parse_size("2M") == 2 * 1024**2
    assert parse_size("1G") == 1 * 1024**3


def test_get_config_defaults():
    """测试 get_config 函数的默认配置，确保在没有配置文件的情况下返回正确的默认值"""
    config = get_config()
    assert config['max_size'] == 5 * 1024 * 1024 * 1024  # 默认 5G
    assert config['log_file'] == ""
    assert config['log_level'] == "INFO"
    assert config['stats'] == False
    assert config['temporary_dir'] == os.path.join(DEFAULT_CACHE_DIR, "tmp")


def test_get_config_with_file():
    """测试 get_config 函数，确保在存在配置文件的情况下能正确读取配置"""
    with tempfile.TemporaryDirectory() as tmpdir:
        cache_dir = os.path.join(tmpdir, 'cache')
        os.makedirs(cache_dir)
        config_file = os.path.join(cache_dir, CONFIG_FILE)
        
        with open(config_file, 'w') as f:
            f.write("""
            max_size=10M
            log_file=test.log
            log_level=DEBUG
            stats=True
            temporary_dir=/tmp/test_tmp
            """)
        
        os.environ['BINCACHE_DIR'] = cache_dir
        
        config = get_config()
        
        assert config['max_size'] == 10 * 1024**2  # 10M
        assert config['log_file'] == os.path.join(cache_dir, 'test.log')
        assert config['log_level'] == "DEBUG"
        assert config['stats'] == True
        assert config['temporary_dir'] == "/tmp/test_tmp"

def test_get_config_with_invalid_data():
    """测试 get_config 函数，确保会忽略配置文件中的错误配置"""
    with tempfile.TemporaryDirectory() as tmpdir:
        cache_dir = os.path.join(tmpdir, 'cache')
        os.makedirs(cache_dir)
        config_file = os.path.join(cache_dir, CONFIG_FILE)
        
        with open(config_file, 'w') as f:
            f.write("""
            max_size
            log_file=test.log
            log_level
            stats=True
            temporary_dir=/tmp/test_tmp
            """)
        
        os.environ['BINCACHE_DIR'] = cache_dir
        
        config = get_config()
        
        assert config['max_size'] == 5 * 1024 * 1024 * 1024 # default
        assert config['log_file'] == os.path.join(cache_dir, 'test.log')
        assert config['log_level'] == "INFO" # default
        assert config['stats'] == True
        assert config['temporary_dir'] == "/tmp/test_tmp"

def test_get_config_with_relative_log_file():
    """测试 get_config 函数，确保在 log_file 配置为相对路径时能正确解析为绝对路径"""
    with tempfile.TemporaryDirectory() as tmpdir:
        cache_dir = os.path.join(tmpdir, 'cache')
        os.makedirs(cache_dir)
        config_file = os.path.join(cache_dir, CONFIG_FILE)
        
        with open(config_file, 'w') as f:
            f.write("""
            log_file=relative/test.log
            """)
        
        os.environ['BINCACHE_DIR'] = cache_dir
        
        config = get_config()
        
        assert config['log_file'] == os.path.join(cache_dir, 'relative', 'test.log')


def test_parse_non_standard_size():
    """测试 parse_size 函数，确保在输入非标准格式时能友好地处理"""
    
    # 没有单位时，直接返回其 int 值
    assert parse_size("1000") == 1000



import os
import tempfile
import pytest
import logging
from unittest import mock
from bincache.logger import get_logger, _logger
from bincache import config as bincache_config
from bincache import logger as bincache_log
import bincache.logger as log_module

# 重置 _logger
@pytest.fixture(autouse=True)
def reset_logger():
    bincache_log._logger = None

def test_get_logger_default_config():
    """测试 get_logger 函数在默认配置下的行为"""
    with mock.patch('bincache.logger.get_config', return_value={
        'log_level': 'INFO',
        'log_file': ''
    }):
        logger = get_logger()
        
        # 检查日志级别
        assert logger.level == logging.INFO
        
        # 检查日志处理程序
        assert len(logger.handlers) == 0


def test_get_logger_with_file():
    """测试 get_logger 函数在提供日志文件的情况下的行为"""
    with tempfile.TemporaryDirectory() as tmpdir:
        log_file = os.path.join(tmpdir, 'test.log')
        
        with mock.patch('bincache.logger.get_config', return_value={
            'log_level': 'DEBUG',
            'log_file': log_file
        }):
            logger = get_logger()
            
            # 检查日志级别
            assert logger.level == logging.DEBUG
            
            # 检查是否有一个日志处理程序
            assert len(logger.handlers) == 1
            
            # 检查日志处理程序的类型
            assert isinstance(logger.handlers[0], logging.FileHandler)
            
            # 检查日志处理程序的文件路径
            assert logger.handlers[0].baseFilename == log_file

            # 检查日志处理程序的格式
            log_formatter = logger.handlers[0].formatter
            assert log_formatter._fmt == '%(asctime)s - PID: %(process)d - PWD: %(pathname)s - %(message)s'


import os
import time
import pickle
import pytest
import tempfile
from unittest import mock
from bincache.cache import get_cache_file_path, put, get, trim_cache_dir_to_limit

@pytest.fixture
def setup_cache_config(monkeypatch, tmpdir):
    temp_cache_dir = str(tmpdir.mkdir("cache"))
    temp_tempdir = str(tmpdir.mkdir("tempdir"))
    mock_config = {'cache_dir': temp_cache_dir, 'temporary_dir': temp_tempdir}
    monkeypatch.setattr('bincache.cache.get_config', lambda: mock_config)
    return mock_config

def test_get_cache_file_path(setup_cache_config):
    key = 'ab1234'
    expected_path = os.path.join(setup_cache_config['cache_dir'], 'ab', '1234')
    actual_path = get_cache_file_path(key)
    assert actual_path == expected_path

def test_put_get(setup_cache_config):
    key = 'ab1234'
    value = {'key': 'value'}
    put(key, value)
    retrieved_value = get(key)
    assert value == retrieved_value

def test_get_with_invalid_key(setup_cache_config):
    key = None
    retrieved_value = get(key)
    assert retrieved_value is None

def test_get_non_existent_key(setup_cache_config):
    key = 'ef9999'
    retrieved_value = get(key)
    assert retrieved_value is None

def test_put_get_small_string(setup_cache_config):
    key = 'small123'
    value = 'a' * 10  # small string
    put(key, value)
    retrieved_value = get(key)
    assert value == retrieved_value

def test_put_get_large_string(setup_cache_config):
    key = 'large123'
    value = 'a' * 10**6  # large string
    put(key, value)
    retrieved_value = get(key)
    assert value == retrieved_value

def test_put_get_binary_content(setup_cache_config):
    key = 'binary123'
    value = b'\x00\xFF\x11\x33\x44'  # binary content
    put(key, value)
    retrieved_value = get(key)
    assert value == retrieved_value

def test_put_get_empty_content(setup_cache_config):
    key = 'empty123'
    value = b''  # empty content
    put(key, value)
    retrieved_value = get(key)
    assert value == retrieved_value

def test_put_get_very_large_content(setup_cache_config):
    key = 'verylarge123'
    value = b'a' * 10**7  # very large binary content
    put(key, value)
    retrieved_value = get(key)
    assert value == retrieved_value

def test_trim_cache_dir_to_limit_no_trim(setup_cache_config):
    key1 = 'key1'
    value1 = 'a' * 100
    key2 = 'key2'
    value2 = 'b' * 100
  
    put(key1, value1)
    put(key2, value2)

    trim_cache_dir_to_limit(setup_cache_config['cache_dir'], 300)

    assert get(key1) == value1
    assert get(key2) == value2

def test_trim_cache_dir_to_limit_with_trim(setup_cache_config):
    key1 = 'key1'
    value1 = 'a' * 100
    key2 = 'key2'
    value2 = 'b' * 100
    key3 = 'key3'
    value3 = 'c' * 100

    put(key1, value1)
    put(key2, value2)
    put(key3, value3)

    os.utime(get_cache_file_path(key1), (time.time() - 3, time.time() - 3))
    os.utime(get_cache_file_path(key2), (time.time() - 2, time.time() - 2))
    os.utime(get_cache_file_path(key3), (time.time() - 1, time.time() - 1))

    trim_cache_dir_to_limit(setup_cache_config['cache_dir'], 200)

    assert get(key1) == None
    assert get(key2) == None
    assert get(key3) == value3

def test_trim_cache_dir_to_limit_with_exact_limit(setup_cache_config):
    key1 = 'key1'
    value1 = 'a' * 100
    size1 = len(pickle.dumps(value1))
    key2 = 'key2'
    value2 = 'b' * 100
    size2 = len(pickle.dumps(value2))
    total_size = size1 + size2

    put(key1, value1)
    put(key2, value2)

    trim_cache_dir_to_limit(setup_cache_config['cache_dir'], total_size)

    assert get(key1) == value1
    assert get(key2) == value2


import hashlib
import subprocess

def hash_file_md5(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as f:
        while True:
            chunk = f.read(8192)
            if not chunk:
                break
            md5.update(chunk)
    return md5.hexdigest()

'''
return list of ('libname, 'libpath', 'address') or None
'''
def get_dynamic_libs(binary):
    result = subprocess.Popen(['ldd', binary], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = result.communicate()
    if result.returncode != 0:
        print(f"Failed to get dynamic libraries for {binary}")
        return None
    libs = []
    for line in stdout.decode('utf-8').splitlines():
        line = line.strip()
        if not line:
            continue
        name, path, address = "", "", ""
        if '(' in line and ')' in line:
            address_start = line.index('(') + 1
            address_end = line.index(')')
            address = line[address_start:address_end].strip()
            line = line[:address_start - 1].strip()
        if '=>' in line:
            parts = line.split('=>')
            name = parts[0].strip()
            path = parts[1].strip()
        elif line.startswith('/'):
            path = line.strip()
        else:
            name = line.strip()
        libs.append((name, path, address))
    return libs

def generate_signature(binary, args):
    if not binary:
        return None
    binary_info = hash_file_md5(binary) 
    libs = get_dynamic_libs(binary)
    if libs is None:
        return None
    libs_info = [(libpath, hash_file_md5(libpath)) for libname, libpath, address in libs if libpath]
    hash_data = str(binary_info) + str(libs_info) + " ".join(args)
    return hashlib.md5(hash_data.encode('utf-8')).hexdigest()


import os
import sys
import subprocess
import shutil

from bincache.cache import get, put
from bincache.signature import generate_signature

def execute_command(argv):
    try:
        command = argv[0]
        result = subprocess.Popen(argv, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = result.communicate()
        stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')
        return result.returncode, stdout, stderr
    except FileNotFoundError:
        return 127, "", f"bincache: command not found: {command}\n"
    except PermissionError:
        return 126, "", f"bincache: permission denied: {command}\n"
    except OSError as e:
        return 1, "", f"bincache: OS error: {command}: {str(e)}\n"
    except Exception as e:
        return 1, "", f"bincache: error executing: {command}: {str(e)}\n"

def main():
    if len(sys.argv) < 2:
        print("Usage: bincache <binary_or_command> <arg1> [arg2 ... argN]")
        print("  <binary_or_command> can be a path to an executable binary or a shell command.")
        print("  <arg1> ... <argN>   are the arguments to be passed to the binary or command.")
        print()
        print("Examples:")
        print("  bincache date")
        print("  bincache ./a.out -l -a")
        sys.exit(1)
    try:
        binary = shutil.which(sys.argv[1])
        args = sys.argv[2:]
        cache_key = generate_signature(binary, args)
        cached_output = get(cache_key)
        if cached_output is not None:
            stdout, stderr = cached_output['stdout'], cached_output['stderr']
            print(stdout, end="")
            print(stderr, end="", file=sys.stderr)
            sys.exit(0)
    except Exception as e:
        pass
    
    returncode, stdout, stderr = execute_command(sys.argv[1:])
    if returncode == 0: # TODO and not stderr:
        try:
            output = {'stdout': stdout, 'stderr': stderr}
            put(cache_key, output)
        except Exception as e:
            pass
    print(stdout, end="")
    print(stderr, end="", file=sys.stderr)
    sys.exit(returncode)

if __name__ == "__main__":
    main()


import os
import tempfile
import hashlib
import pytest
import subprocess
from unittest import mock
from bincache.signature import hash_file_md5, get_dynamic_libs, generate_signature

def test_hash_file_md5():
    content = b"test content for hashing"
    expected_md5 = hashlib.md5(content).hexdigest()
    
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        tmpfile.write(content)
        tmpfile_path = tmpfile.name
    
    try:
        calculated_md5 = hash_file_md5(tmpfile_path)
        assert calculated_md5 == expected_md5
    finally:
        os.remove(tmpfile_path)

def test_get_dynamic_libs():
    example_output = b"""
    linux-vdso.so.1 =>  (0x00007fff6ab93000)
	libstdc++.so.6 => /lib64/libstdc++.so.6 (0x00007fc5c64c6000)
	libm.so.6 => /lib64/libm.so.6 (0x00007fc5c61c4000)
	libgcc_s.so.1 => /lib64/libgcc_s.so.1 (0x00007fc5c5fad000)
	libc.so.6 => /lib64/libc.so.6 (0x00007fc5c5bdf000)
	/lib64/ld-linux-x86-64.so.2 (0x00007fc5c6853000)
    """
    with mock.patch('subprocess.Popen') as mock_popen:
        process_mock = mock.Mock()
        attrs = {'communicate.return_value': (example_output, b''), 'returncode': 0}
        process_mock.configure_mock(**attrs)
        mock_popen.return_value = process_mock
        
        libs = get_dynamic_libs('dummy_binary')

        expected_libs = [
            ('linux-vdso.so.1', '', '0x00007fff6ab93000'),
            ('libstdc++.so.6', '/lib64/libstdc++.so.6', '0x00007fc5c64c6000'),
            ('libm.so.6', '/lib64/libm.so.6', '0x00007fc5c61c4000'),
            ('libgcc_s.so.1', '/lib64/libgcc_s.so.1', '0x00007fc5c5fad000'),
            ('libc.so.6', '/lib64/libc.so.6', '0x00007fc5c5bdf000'),
            ('', '/lib64/ld-linux-x86-64.so.2', '0x00007fc5c6853000')
        ]
        
        assert libs == expected_libs
@mock.patch('bincache.signature.hash_file_md5', autospec=True)
@mock.patch('bincache.signature.get_dynamic_libs', autospec=True)
def test_generate_signature(mock_get_dynamic_libs, mock_hash_file_md5):
    # 定义根据文件路径返回不同的哈希值的逻辑
    def hash_side_effect(path):
        hash_dict = {
            'dummy_binary': 'hash_of_dummy_binary',
            '/lib64/libstdc++.so.6': 'hash_of_libstdc++',
            '/lib64/libm.so.6': 'hash_of_libm',
            '/lib64/libgcc_s.so.1': 'hash_of_libgcc_s',
            '/lib64/libc.so.6': 'hash_of_libc',
            '/lib64/ld-linux-x86-64.so.2': 'hash_of_ld-linux-x86-64.so.2'
        }
        return hash_dict.get(path, 'hash_of_unknown_lib')
    
    mock_hash_file_md5.side_effect = hash_side_effect
    
    # 模拟 get_dynamic_libs 返回的动态库信息
    mock_get_dynamic_libs.return_value = [
        ('linux-vdso.so.1', '', '0x00007fff6ab93000'),
        ('libstdc++.so.6', '/lib64/libstdc++.so.6', '0x00007fc5c64c6000'),
        ('libm.so.6', '/lib64/libm.so.6', '0x00007fc5c61c4000'),
        ('libgcc_s.so.1', '/lib64/libgcc_s.so.1', '0x00007fc5c5fad000'),
        ('libc.so.6', '/lib64/libc.so.6', '0x00007fc5c5bdf000'),
        ('', '/lib64/ld-linux-x86-64.so.2', '0x00007fc5c6853000')
    ]
    
    binary = 'dummy_binary'
    args = ['arg1', 'arg2']
    
    # 计算生成的签名应包含的哈希值和信息
    expected_binary_hash = 'hash_of_dummy_binary'
    expected_libs_info = [
        ('/lib64/libstdc++.so.6', 'hash_of_libstdc++'),
        ('/lib64/libm.so.6', 'hash_of_libm'),
        ('/lib64/libgcc_s.so.1', 'hash_of_libgcc_s'),
        ('/lib64/libc.so.6', 'hash_of_libc'),
        ('/lib64/ld-linux-x86-64.so.2', 'hash_of_ld-linux-x86-64.so.2')
    ]
    expected_signature_source = f"{expected_binary_hash}{expected_libs_info}arg1 arg2"
    expected_signature = hashlib.md5(expected_signature_source.encode('utf-8')).hexdigest()
    
    # 调用 generate_signature 函数生成签名
    signature = generate_signature(binary, args)
    
    # 验证生成的签名是否正确
    assert signature == expected_signature


import logging
from bincache.config import get_config

_logger = None

def get_logger():
    global _logger
    if _logger is None:
        config = get_config()
        _logger = logging.getLogger('bincache')
        log_level = getattr(logging, config['log_level'], logging.INFO)
        _logger.setLevel(log_level)
        if config['log_file']:
            log_handler = logging.FileHandler(config['log_file'])
            log_formatter = logging.Formatter('%(asctime)s - PID: %(process)d - PWD: %(pathname)s - %(message)s')
            log_handler.setFormatter(log_formatter)
            _logger.addHandler(log_handler)
    return _logger

def debug(msg):
    get_logger().debug(msg)

def info(msg):
    get_logger().info(msg)

def warning(msg):
    get_logger().warning(msg)

def error(msg):
    get_logger().error(msg)

def critical(msg):
    get_logger().critical(msg)


import os
import sys
import subprocess
import shutil
import pytest
from unittest import mock
from io import StringIO

from bincache.cli import main
from bincache.cache import get, put
from bincache.config import get_config

@pytest.fixture
def mock_config(monkeypatch, tmpdir):
    temp_cache_dir = str(tmpdir.mkdir("cache"))
    temp_tempdir = str(tmpdir.mkdir("tempdir"))
    mock_config = {'cache_dir': temp_cache_dir, 'temporary_dir': temp_tempdir}
    monkeypatch.setattr('bincache.config.get_config', lambda: mock_config)
    return mock_config

@pytest.fixture
def mock_binary(monkeypatch):
    alias_map = {
        'dummy_command_cache': '/bin/dummy_command',
        'echo': '/bin/echo',
        './error': './error',
        './error_but_cache': './error_but_cache',
        './command_with_stderr': './command_with_stderr',
    }
    cache_map = {
        ('/bin/dummy_command',): {
            'signature': 'dummy_command_signature',
            'cached_output': 'dummy_command_cached_output',
        },
        ('/bin/echo', 'Hello'): {
            'signature': 'echo_signature',
        },
        ('./error',): {
            'signature': 'error_signature',
        },
        ('./error_but_cache',): {
            'signature': 'error_but_cache_signature',
            'cached_output': 'error_but_cached_output',
        },
        ('./command_with_stderr',): {
            'signature': 'command_with_stderr_signature',
        },
    }
    command_map = {
        ('dummy_command_cache',): {
            'stdout': 'dummy_command_stdout',
            'stderr': '',
            'returncode': 0
        },
        ('echo', 'Hello'): {
            'stdout': 'Hello',
            'stderr': '',
            'returncode': 0
        },
        ('./error',): {
            'stdout': 'error_stdout',
            'stderr': 'error_stderr',
            'returncode': 1
        },
        ('./error_but_cache',): {
            'stdout': 'error_stdout',
            'stderr': 'error_stderr',
            'returncode': 1
        },
        ('./command_with_stderr',): {
            'stdout': 'error_stdout',
            'stderr': 'error_stderr',
            'returncode': 0
        },
        ('./not_found',): {
            'raise_error': FileNotFoundError
        },
        ('./no_permission',): {
            'raise_error': PermissionError
        },
        ('./os_error',): {
            'raise_error': OSError
        },
        ('./other_error',): {
            'raise_error': Exception
        }
    }

    def which(cmd):
        return alias_map.get(cmd)

    def generate_signature(binary, args):
        details = cache_map.get((binary,) + tuple(args))
        return details.get('signature') if details else None

    def cache_get(key):
        if key is None:
            return None
        for binary, details in cache_map.items():
            if key == details.get('signature') and 'cached_output' in details:
                return {'stdout': details.get('cached_output'), 'stderr': ''}
        return None

    def popen_mock(*popen_args, **kwargs):
        cmd = popen_args[0]
        details = command_map.get(tuple(cmd))
        if details:
            if 'raise_error' in details:
                raise details['raise_error']
            mock_proc = mock.Mock()
            mock_proc.communicate.return_value = (details['stdout'].encode('utf-8'), details['stderr'].encode('utf-8'))
            mock_proc.returncode = details.get('returncode')
            return mock_proc
        raise ValueError(f"Unexpected command: {cmd}")
    monkeypatch.setattr('bincache.cli.shutil.which', which)
    monkeypatch.setattr('bincache.cli.generate_signature', generate_signature)
    monkeypatch.setattr('bincache.cli.get', cache_get)
    monkeypatch.setattr(subprocess, 'Popen', popen_mock)

# case: 没有提供命令行参数
def test_no_arguments(monkeypatch, capsys):
    monkeypatch.setattr(sys, 'argv', ['bincache'])
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    captured = capsys.readouterr()
    assert "Usage: bincache <binary_or_command> <arg1> [arg2 ... argN]" in captured.out

# case: 命中缓存
def test_cached_output(monkeypatch, mock_binary, capsys):
    monkeypatch.setattr(sys, 'argv', ['bincache', 'dummy_command_cache'])
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0
    captured = capsys.readouterr()
    assert captured.out == "dummy_command_cached_output"

# case: 未命中缓存但执行成功，并验证结果被缓存
def test_exec_command_and_cache(monkeypatch, mock_config, capsys, mock_binary):
    monkeypatch.setattr(sys, 'argv', ['bincache', 'echo', 'Hello'])

    put_mock = mock.Mock()
    monkeypatch.setattr('bincache.cli.put', put_mock)
    
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0

    captured = capsys.readouterr()
    assert captured.out.strip() == "Hello"
    
    # Verify that the command output was cached
    put_mock.assert_called_once_with('echo_signature', {'stdout': 'Hello', 'stderr': ''})

# case: 未命中缓存且执行失败，并验证结果不会被缓存
def test_exec_command_with_error(monkeypatch, mock_config, capsys, mock_binary):
    monkeypatch.setattr(sys, 'argv', ['bincache', './error'])

    put_mock = mock.Mock()
    monkeypatch.setattr('bincache.cli.put', put_mock)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    captured = capsys.readouterr()
    assert captured.out.strip() == 'error_stdout'
    assert captured.err.strip() == 'error_stderr'

    # Verify that the command output was not cached
    put_mock.assert_not_called()

# case: 命中缓存，哪怕实际结果可能会失败，也会返回缓存
def test_cached_output_with_error(monkeypatch, mock_binary, capsys):
    monkeypatch.setattr(sys, 'argv', ['bincache', './error_but_cache'])
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0  # Assuming we read from cache and do not execute
    captured = capsys.readouterr()
    assert captured.out == "error_but_cached_output"

# case: 未命中缓存且执行成功，即使有stderr，也会被缓存
def test_exec_command_with_error(monkeypatch, mock_config, capsys, mock_binary):
    monkeypatch.setattr(sys, 'argv', ['bincache', './command_with_stderr'])

    put_mock = mock.Mock()
    monkeypatch.setattr('bincache.cli.put', put_mock)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0

    captured = capsys.readouterr()
    assert captured.out.strip() == 'error_stdout'
    assert captured.err.strip() == 'error_stderr'

    # Verify that the command output was cached
    put_mock.assert_called_once_with('command_with_stderr_signature', {'stdout': 'error_stdout', 'stderr': 'error_stderr'})

# case: 命令不存在
def test_exec_not_found(monkeypatch, mock_binary, mock_config, capsys):
    monkeypatch.setattr(sys, 'argv', ['bincache', './not_found'])

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 127

    captured = capsys.readouterr()
    assert captured.out.strip() == ''
    assert captured.err.strip() == 'bincache: command not found: ./not_found'

# case: 命令无权执行
def test_exec_no_permission(monkeypatch, mock_binary, mock_config, capsys):
    monkeypatch.setattr(sys, 'argv', ['bincache', './no_permission'])

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 126

    captured = capsys.readouterr()
    assert captured.out.strip() == ''
    assert captured.err.strip() == 'bincache: permission denied: ./no_permission'

# case: os错误
def test_exec_os_error(monkeypatch, mock_binary, mock_config, capsys):
    monkeypatch.setattr(sys, 'argv', ['bincache', './os_error'])

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    captured = capsys.readouterr()
    assert captured.out.strip() == ''
    assert 'bincache: OS error: ./os_error:' in captured.err.strip()

# case: 其他错误 
def test_exec_other_error(monkeypatch, mock_binary, mock_config, capsys):
    monkeypatch.setattr(sys, 'argv', ['bincache', './other_error'])

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

    captured = capsys.readouterr()
    assert captured.out.strip() == ''
    assert 'bincache: error executing: ./other_error:' in captured.err.strip()

# case: 即使bincache内部异常，也会正常执行命令
def test_exec_command_with_bincache_error(monkeypatch, mock_config, capsys, mock_binary):
    monkeypatch.setattr(sys, 'argv', ['bincache', 'echo', 'Hello'])

    def raise_error(*args, **kwargs):
        raise Exception('bincache error')

    monkeypatch.setattr('bincache.cli.shutil.which', raise_error)
    monkeypatch.setattr('bincache.cli.generate_signature', raise_error)
    monkeypatch.setattr('bincache.cli.put', raise_error)
    monkeypatch.setattr('bincache.cli.get', raise_error)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0

    captured = capsys.readouterr()
    assert captured.out.strip() == "Hello"

# case: 即使bincache内部全部返回None，也会正常执行命令
def test_exec_command_with_bincache_error(monkeypatch, mock_config, capsys, mock_binary):
    monkeypatch.setattr(sys, 'argv', ['bincache', 'echo', 'Hello'])

    def return_None(*args, **kwargs):
        return None

    monkeypatch.setattr('bincache.cli.shutil.which', return_None)
    monkeypatch.setattr('bincache.cli.generate_signature', return_None)
    monkeypatch.setattr('bincache.cli.put', return_None)
    monkeypatch.setattr('bincache.cli.get', return_None)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        main()
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0

    captured = capsys.readouterr()
    assert captured.out.strip() == "Hello"


import os
import pickle
import tempfile
from bincache.config import get_config
from bincache import logger

def get_cache_file_path(key):
    config = get_config()
    prefix = key[:2]
    filename = key[2:]
    cache_file_folder = os.path.join(config['cache_dir'], prefix)
    return os.path.join(cache_file_folder, filename)

# TODO Add file locking or other mechanism to handle multi-process access
def trim_cache_dir_to_limit(cache_dir, limit_size_in_bytes):
    """removing old files if the total size exceeds the limit."""
    total_size = 0
    file_paths = []
    for root, dirs, files in os.walk(cache_dir):
        # Ignore the files in the first level directory (configuration files)
        if root == cache_dir:
            continue
        for name in files:
            file_path = os.path.join(root, name)
            file_size = os.path.getsize(file_path)
            total_size += file_size
            file_paths.append((file_path, os.path.getmtime(file_path), file_size))
    logger.info(f"totle_size: {total_size}, limit_size_in_bytes: {limit_size_in_bytes}")
    if total_size <= limit_size_in_bytes:
        return
    # Sort files by modification time (oldest first)
    file_paths.sort(key=lambda x: x[1])
    for file_path, _, file_size in file_paths:
        os.remove(file_path)
        logger.info(f"{file_path} removed, totle_size: {total_size}, limit_size_in_bytes: {limit_size_in_bytes}")
        total_size -= file_size
        if total_size <= limit_size_in_bytes:
            break

def put(key, value):
    config = get_config()
    cache_file_path = get_cache_file_path(key)
    if cache_file_path is None:
        return
    cache_file_folder = os.path.dirname(cache_file_path)
    os.makedirs(cache_file_folder, exist_ok=True)
    if config['temporary_dir']:
        os.makedirs(config['temporary_dir'], exist_ok=True)
        with tempfile.NamedTemporaryFile(delete=False, dir=config['temporary_dir']) as temp_file:
            temp_path = temp_file.name
            pickle.dump(value, temp_file)
        try:
            os.rename(temp_path, cache_file_path)
        finally:
            if os.path.exists(temp_path):
                os.remove(temp_path)
    else:
        pickle.dump(value, cache_file_path)

    if 'max_size' in config:
        trim_cache_dir_to_limit(config['cache_dir'], config['max_size'])

def get(key):
    if not key:
        return None
    cache_file_path = get_cache_file_path(key)
    if not cache_file_path:
        return None
    try:
        with open(cache_file_path, 'rb') as f:
            data = f.read()
            return pickle.loads(data)
    except FileNotFoundError:
        return None


from setuptools import setup, find_packages

setup(
    name="bincache",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'bincache = bincache.cli:main',
        ],
    },
    author="simpx",
    author_email="simpxx@gmail.com",
    description="Bincache is a command-line utility designed to cache the output of executable binaries and shell commands.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/simpx/bincache",
    license='MIT',
    install_requires=[],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    tests_require=[
        'pytest',
        'pytest-mock',
    ],
    test_suite='tests',
)
